# Copyright (c) 2020-2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import math
import numpy

import btype

from .hex_funcs import hex_to_double


SUPERSCRIPT = [
    '\u2070',
    '\u00B9',
    '\u00B2',
    '\u00B3',
    '\u2074',
    '\u2075',
    '\u2076',
    '\u2077',
    '\u2078',
    '\u2079',
    ]

SUBSCRIPT = [
    '\u2080',
    '\u2081',
    '\u2082',
    '\u2083',
    '\u2084',
    '\u2085',
    '\u2086',
    '\u2087',
    '\u2088',
    '\u2089',
    ]

TYPE_TABLE = {}
TYPE_TABLE_DOUBLE = {}


def number_script(v, script_array):
    assert isinstance(v, int)

    v = str(v)
    s = ''
    for c in v:
        s += script_array[ord(c) - 48]
    return s


def superscript(v):
    if v == 1:
        return ''
    return number_script(v, SUPERSCRIPT)


def subscript(v):
    return number_script(v, SUBSCRIPT)


class PolyWindow(btype.Struct):
    low     = btype.uint64_t(0xFFFFFFFFFFFFFFFF)
    high    = btype.uint64_t(0xFFFFFFFFFFFFFFFF)


def make_poly_bstruct(Nvars, Ncoefs):
    if (Nvars, Ncoefs) in TYPE_TABLE:
        return TYPE_TABLE[(Nvars, Ncoefs)]

    class _Poly(btype.Struct):
        nvars   = btype.uint32_t(0xFFFFFFFF)
        order   = btype.uint32_t(0xFFFFFFFF)
        windows = btype.Array(PolyWindow(), Nvars)
        coefs   = btype.Array(btype.uint64_t(0xFFFFFFFFFFFFFFFF), Ncoefs)
    TYPE_TABLE[(Nvars, Ncoefs)] = _Poly

    return _Poly


class PolynomialFit1D:
    def __init__(self, order, pf):
        self.order = order
        self.pf    = pf

    @staticmethod
    def from_domain_coefs(x_domain, coefs):
        '''
        Given the x_domain and a list of coefficients of the form:

            [x0, x1, x2, ..., xN]

        generate the fit polynomial for evaluation purposes.
        '''
        order = len(coefs) - 1
        pf    = numpy.polynomial.polynomial.Polynomial(coefs, domain=x_domain)
        return PolynomialFit1D(order, pf)

    @staticmethod
    def from_poly_bstruct(p):
        '''
        Generate the fit polynomial from a make_poly_bstruct() btype class.
        '''
        assert p.nvars == 1
        order  = p.order
        x_domain = (hex_to_double(p.windows[0].low),
                    hex_to_double(p.windows[0].high))
        coefs = [hex_to_double(p.coefs[i]) for i in range(order + 1)]
        return PolynomialFit1D.from_domain_coefs(x_domain, coefs)

    def __call__(self, x):
        return self.pf(x)

    @staticmethod
    def _polystr(coef):
        s      = '%.5f' % coef[0]
        for i, c in enumerate(coef[1:]):
            s += ' + %.5f*x%s' % (c, superscript(i + 1))
        return s

    def __repr__(self):
        s = PolynomialFit1D._polystr(self.pf.coef)
        s += '\n : x(%s, %s) -> [-1. 1.]' % (float(self.pf.domain[0]),
                                             float(self.pf.domain[1]))
        return s


class PolynomialFit2D:
    def __init__(self, x_domain, y_domain, coefs):
        self.x_domain = x_domain
        self.y_domain = y_domain
        self.coefs    = coefs

    @staticmethod
    def from_domain_coefs(x_domain, y_domain, coefs):
        '''
        Given the x_domain, y_domain and a list of coefficients of the form:

            [x0y0, x1y0, x2y0, ..., xNy0,
             x0y1, x1y1, x2y1, ..., xNy1,
             ...
             x0yN, x1yN, x2yN, ..., xNyN]

        generate the fit polynomial for evaluation purposes.
        '''
        order = round(math.sqrt(len(coefs))) - 1
        return PolynomialFit2D(x_domain, y_domain,
                               numpy.reshape(coefs, (order + 1, order + 1)))

    @staticmethod
    def from_poly_bstruct(p):
        '''
        Generate the fit polynomial from a make_poly_bstruct() btype class.
        '''
        assert p.nvars == 2
        order = p.order
        x_domain = (hex_to_double(p.windows[0].low),
                    hex_to_double(p.windows[0].high))
        y_domain = (hex_to_double(p.windows[1].low),
                    hex_to_double(p.windows[1].high))
        coefs = [hex_to_double(p.coefs[i]) for i in range((order + 1)**2)]
        return PolynomialFit2D.from_domain_coefs(x_domain, y_domain, coefs)

    @staticmethod
    def _mapped(p, domain):
        l = domain[0]
        h = domain[1]
        return 2*(p - l)/(h - l) - 1

    def __call__(self, x, y):
        x = self._mapped(x, self.x_domain)
        y = self._mapped(y, self.y_domain)
        return numpy.polynomial.polynomial.polyval2d(x, y, self.coefs)

    @staticmethod
    def _polystr(coef):
        deg_x, deg_y = coef.shape
        s = '   |'
        for x in range(deg_x):
            if x == 0:
                p = '1'
            else:
                p = 'x%s' % superscript(x)
            s += ' %12s' % p
        l = '---+'
        for x in range(deg_x):
            l += ' ------------'
        s += '\n%s\n' % l
        for y in range(deg_y):
            if y == 0:
                p = '1'
            else:
                p = 'y%s' % superscript(y)
            s += '%-2s |' % p
            for x in range(deg_x):
                s += ' %12.5f' % coef[x][y]
            s += '\n'
        s += l
        return s

    def __repr__(self):
        return PolynomialFit2D._polystr(self.coefs) + (
            '\n : x%s -> [-1. 1.]\n : y%s -> [-1. 1.]' %
            (self.x_domain, self.y_domain))



#!/usr/bin/env python3
# Copyright (c) 2024 by Phase Advanced Sensor Systems, Inc.
# All rights reserved.
import threading
import argparse
import logging
import time

from xtalx.tools.config import Config
from xtalx.tools.influxdb import InfluxDBPushQueue
import xtalx.p_sensor
import xtalx.spi_adapter


# Verbosity
LOG_LEVEL = logging.INFO


def sensor_thread(xhtiss, idb):
    # Monitor the sensor.
    logging.info('%s: Monitoring...', xhtiss.serial_num)
    for m in xhtiss.yield_measurements(poll_interval_sec=0.05):
        if idb:
            point = m.to_influx_point()
            if point['fields']:
                idb.append(point)

        logging.info('%s: t %.3f p %.3f tf %.6f pf %.6f ',
                     xhtiss.serial_num, m.temp_c, m.pressure_psi,
                     m.temp_freq, m.pressure_freq)


def main(rv):
    # Read the configuration file.
    if rv.config:
        logging.info('Reading configuration...')
        with open(rv.config, encoding='utf8') as f:
            c = Config(f.readlines(), ['influx_host', 'influx_user',
                                       'influx_password', 'influx_database'])

        # Open a connection to InfluxDB.
        logging.info('Connecting to InfluxDB...')
        idb = InfluxDBPushQueue(c.influx_host, 8086, c.influx_user,
                                c.influx_password, database=c.influx_database,
                                ssl=True, verify_ssl=True,
                                timeout=100, throttle_secs=10)
    else:
        idb = None

    # Make the bus.
    dev = xtalx.spi_adapter.find_one_spia(
            serial_number=rv.adapter_serial_number)
    if dev is None:
        raise Exception('No adapter found.')
    bus = xtalx.spi_adapter.make_spia(dev)
    bus.set_vext(True)
    time.sleep(0.1)

    # Spawn a thread for each sensor on the bus.
    xhtiss = xtalx.p_sensor.XHTISS(bus)
    logging.info('%s: Found sensor with firmware version %s, git SHA1 %s',
                 xhtiss.serial_num, xhtiss.fw_version_str, xhtiss.git_sha1)
    t_c, p_c, sample_ms = xhtiss.get_flash_params()
    logging.info('%s: T Coefficient: %u', xhtiss.serial_num, t_c)
    logging.info('%s: P Coefficient: %u', xhtiss.serial_num, p_c)
    logging.info('%s: Sample Interval: %u ms', xhtiss.serial_num, sample_ms)

    t = threading.Thread(target=sensor_thread, args=(xhtiss, idb))
    t.start()

    # Run forever.
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print()

    # Halt and join all threads for a clean exit.
    xhtiss.halt_yield()
    t.join()


def _main():
    logging.basicConfig(format='\033[1m[%(asctime)s.%(msecs)03d]\033[0m '
                        '%(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    logging.getLogger().setLevel(LOG_LEVEL)

    parser = argparse.ArgumentParser()
    parser.add_argument('--config', '-c')
    parser.add_argument('--adapter-serial-number', '-s')

    try:
        main(parser.parse_args())
    except KeyboardInterrupt:
        print()


if __name__ == '__main__':
    _main()
